"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});

exports.default = function (babel) {
  var t = babel.types;

  (0, _babelInstance.setTypes)(t);

  // HELPER FUNCTIONS
  function bindMethods(path, methodIds) {
    var assignId = void 0,
        inExpression = false;
    if (path.isClassDeclaration()) {
      assignId = path.node.id;
    } else if (path.parentPath.isAssignmentExpression() && path.parentPath.parentPath.isExpressionStatement()) {
      assignId = path.parentPath.node.left;
    } else if (path.parentPath.isVariableDeclarator()) {
      assignId = path.parentPath.node.id;
    } else {
      var id = path.isClass() ? "class" : "obj";
      assignId = path.getStatementParent().scope.generateDeclaredUidIdentifier(id);
      inExpression = true;
    }
    t.assertOneOf(assignId, ["Identifier", "MemberExpression"]);

    var assignments = methodIds.map(function (methodId) {
      // could be computed, eg `['blah']() => {}`
      t.assertOneOf(methodId, ["Identifier", "Expression"]);
      var isComputed = !t.isIdentifier(methodId);
      var objDotMethod = t.memberExpression(assignId, methodId, isComputed);
      var bind = t.callExpression(t.memberExpression(objDotMethod, t.identifier("bind")), [assignId]);
      return t.assignmentExpression("=", objDotMethod, bind);
    });

    if (inExpression) {
      path.replaceWith(t.sequenceExpression([t.assignmentExpression("=", assignId, path.node)].concat(_toConsumableArray(assignments), [assignId])));
    } else {
      path.getStatementParent().insertAfter(assignments.map(function (a) {
        return t.expressionStatement(a);
      }));
    }
  }

  function shouldParseAsLightScript(file) {
    if (!file || !file.opts || !file.opts.filename) return true;
    var filename = file.opts.filename;
    // HACK: for lightscript-eslint, and possibly others

    if (filename === "unknown") return true;

    // TODO: consider "peeking" at the first line for a shebang or 'use lightscript' directive.
    return (
      // HACK: allow parsing .js test files in this repo.
      // TODO: modify `babel-helper-plugin-test-runner` or something instead
      filename.includes("test/fixtures") || filename.includes(".lsc") || filename.includes(".lsx")
    );
  }

  // eg; 'react', 'lodash/fp', './actions'


  // eg; "React", "PropTypes"


  function initializeStdlib(opts) {
    if (opts.stdlib === false) return false;

    if (typeof opts.stdlib === "object") {
      return Object.assign({}, opts.stdlib.lodash === false ? {} : _stdlib.lodashImports);
    }

    return _stdlib.defaultImports;
  }

  function collectStdlibImport(stdlib, imports, specifier) {
    if (!stdlib) return;

    var importPath = stdlib[specifier];

    if (!imports[importPath]) {
      imports[importPath] = [];
    }

    if (imports[importPath].indexOf(specifier) < 0) {
      imports[importPath].push(specifier);
    }
  }

  function insertStdlibImports(path, imports, useRequire) {
    var declarations = [];
    for (var importPath in imports) {
      var specifierNames = imports[importPath];
      var specifiers = [];

      if (useRequire) {
        // eg; `const { map, uniq } = require('lodash');`
        var _iteratorNormalCompletion = true;
        var _didIteratorError = false;
        var _iteratorError = undefined;

        try {
          for (var _iterator = specifierNames[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
            var specifierName = _step.value;

            var importIdentifier = t.identifier(specifierName);
            specifiers.push(t.objectProperty(importIdentifier, importIdentifier, false, true));
          }
        } catch (err) {
          _didIteratorError = true;
          _iteratorError = err;
        } finally {
          try {
            if (!_iteratorNormalCompletion && _iterator.return) {
              _iterator.return();
            }
          } finally {
            if (_didIteratorError) {
              throw _iteratorError;
            }
          }
        }

        var requirePattern = t.objectPattern(specifiers);
        var requireCall = t.callExpression(t.identifier("require"), [t.stringLiteral(importPath)]);
        var requireStmt = t.variableDeclaration("const", [t.variableDeclarator(requirePattern, requireCall)]);
        declarations.push(requireStmt);
      } else {
        // eg; `import { map, uniq } from 'lodash';`
        var _iteratorNormalCompletion2 = true;
        var _didIteratorError2 = false;
        var _iteratorError2 = undefined;

        try {
          for (var _iterator2 = specifierNames[Symbol.iterator](), _step2; !(_iteratorNormalCompletion2 = (_step2 = _iterator2.next()).done); _iteratorNormalCompletion2 = true) {
            var _specifierName = _step2.value;

            var _importIdentifier = t.identifier(_specifierName);
            specifiers.push(t.importSpecifier(_importIdentifier, _importIdentifier));
          }
        } catch (err) {
          _didIteratorError2 = true;
          _iteratorError2 = err;
        } finally {
          try {
            if (!_iteratorNormalCompletion2 && _iterator2.return) {
              _iterator2.return();
            }
          } finally {
            if (_didIteratorError2) {
              throw _iteratorError2;
            }
          }
        }

        var importDeclaration = t.importDeclaration(specifiers, t.stringLiteral(importPath));
        declarations.push(importDeclaration);
      }
    }
    path.unshiftContainer("body", declarations);
  }

  // traverse as top-level item so as to run before other babel plugins
  // (and avoid traversing any of their output)
  function Program(path, state) {
    if (!shouldParseAsLightScript(state.file)) return;

    var stdlib = initializeStdlib(state.opts);
    var inlinedOperatorsEnabled = (0, _inlinedOperators.getInlinedOperatorsEnabled)(state.opts);
    var useRequire = state.opts.stdlib && state.opts.stdlib.require === true;
    var imports = {};

    path.traverse({

      ForInArrayStatement(path) {
        (0, _for.transformForInArrayStatement)(path);
      },

      ForInObjectStatement(path) {
        (0, _for.transformForInObjectStatement)(path);
      },

      ArrayComprehension(path) {
        (0, _comprehension.transformArrayComprehension)(path);
      },

      ObjectComprehension(path) {
        (0, _comprehension.transformObjectComprehension)(path);
      },

      TildeCallExpression: {
        // run on exit instead of enter so that SafeMemberExpression
        // can process differently from a wrapping CallExpression
        // eg; `a?.b~c()` -> `a == null ? null : c(a.b)`
        exit(path) {
          var args = [path.node.left].concat(_toConsumableArray(path.node.arguments));
          var callExpr = t.callExpression(path.node.right, args);

          if (inlinedOperatorsEnabled && (0, _inlinedOperators.replaceWithInlinedOperator)(path, path.node.right, args)) return;
          if (path.node.safe) {
            (0, _safeCall.replaceWithSafeCall)(path, callExpr);
          } else {
            path.replaceWith(callExpr);
          }
        }
      },

      CallExpression: {
        exit(path) {
          if (inlinedOperatorsEnabled && (0, _inlinedOperators.replaceWithInlinedOperator)(path, path.node.callee, path.node.arguments)) return;
          if (path.node.safe) {
            (0, _safeCall.replaceWithSafeCall)(path, path.node);
          }
        }
      },

      NamedArrowFunction(path) {
        if (path.node.skinny) {
          (0, _functions.replaceWithPlainFunction)(path);
        } else if (path.node.generator) {
          // there are no arrow-generators in ES6, so can't compile to arrow
          (0, _functions.replaceWithBoundFunction)(path);
        } else {
          (0, _functions.replaceWithArrowFunction)(path);
        }
      },

      NamedArrowMemberExpression(path) {
        var object = path.node.object;
        var node = path.node;
        delete node.object;
        node.type = "NamedArrowExpression";

        if (!node.skinny) {
          node.skinny = true; // binding here, don't turn into arrow
          node = t.callExpression(t.memberExpression(node, t.identifier("bind")), [object]);
        }

        path.replaceWith(t.assignmentExpression("=", t.memberExpression(object, path.node.id), node));
      },

      ArrowFunctionExpression(path) {
        if (path.node.skinny) {
          (0, _functions.replaceWithPlainFunction)(path);
        } else if (path.node.generator) {
          (0, _functions.replaceWithBoundFunction)(path);
        }
      },

      Method(path) {
        if ((0, _functions.isNamedArrowFunction)(path.node)) {
          path.node.body = (0, _blocks.toBlockStatement)(path.node.body);
        }
      },

      ClassBody(path) {
        var fatArrows = [],
            fatStaticArrows = [],
            constructorPath = void 0;
        path.node.body.forEach(function (method, i) {
          if (!t.isMethod(method)) return;

          if (method.kind === "constructor") {
            constructorPath = path.get(`body.${i}.body`);
          } else if (method.static && method.skinny === false) {
            fatStaticArrows.push(method.key);
            method.skinny = true; // prevent infinite recursion
          } else if (method.skinny === false) {
            fatArrows.push(method.key);
            method.skinny = true; // prevent infinite recursion
          }
        });

        var maybeAddSuper = path.parentPath.node.superClass && constructorPath;
        if (fatArrows.length || maybeAddSuper) {
          constructorPath = (0, _classes.ensureConstructor)(path.parentPath, constructorPath, true);
        }

        if (fatArrows.length) {
          (0, _classes.bindMethodsInConstructor)(path.parentPath, constructorPath, fatArrows);
        }

        if (fatStaticArrows.length) {
          bindMethods(path.parentPath, fatStaticArrows);
        }
      },

      ObjectExpression(path) {
        var fatArrows = [];
        path.node.properties.forEach(function (prop) {
          if (t.isMethod(prop) && prop.skinny === false) {
            fatArrows.push(prop.key);
            // bit ugly, but need a way to ensure we don't double-recurse...
            prop.skinny = true;
          }
        });

        if (fatArrows.length) {
          bindMethods(path, fatArrows);
        }
      },

      Function: {
        exit(path) {
          (0, _functions.addImplicitReturns)(path);

          // As this is an exit visitor, other LSC transforms have reduced
          // arrows to plain FunctionDeclarations by this point.
          if (path.node.type === "FunctionDeclaration") {
            // somehow this wasn't being done... may signal deeper issues...
            path.getFunctionParent().scope.registerDeclaration(path);
          }
        }
      },

      IfExpression(path) {
        var consequent = (0, _blocks.blockToExpression)(path.get("consequent"));

        var alternate = void 0;
        if (path.node.alternate) {
          alternate = (0, _blocks.blockToExpression)(path.get("alternate"));
        } else {
          alternate = t.nullLiteral();
        }

        path.replaceWith(t.conditionalExpression(path.node.test, consequent, alternate));
      },

      AssignmentExpression(path) {
        if (path.node.operator === "<-" || path.node.operator === "<!-") {
          path.node.operator = "=";
        }

        // TODO: consider enforcing `now` for MemberExpression too
        if ((0, _is2.default)("MemberExpression", path.node.left)) return;

        if (path.node.isNowAssign === false) {
          throw path.buildCodeFrameError("Incorrect assignment: to reassign, use `now`; to assign as `const`, put on its own line.");
        }
      },

      SafeAwaitExpression(path) {
        var errId = path.scope.generateUidIdentifier("err");

        var tryCatch = t.tryStatement(t.blockStatement([t.returnStatement(t.awaitExpression(path.node.argument))]), t.catchClause(errId, t.blockStatement([t.returnStatement(errId)])));
        var fn = t.arrowFunctionExpression([], t.blockStatement([tryCatch]), true);
        // TODO: returntype annotation
        var iife = t.callExpression(fn, []);
        var awaitExpr = t.awaitExpression(iife);
        path.replaceWith(awaitExpr);
      },

      SafeMemberExpression(path) {
        // x?.y -> x == null ? x : x.y
        // x?[y] -> x == null ? x : x[y]
        var node = path.node;
        var object = node.object;


        var left = void 0;
        if (object.type === "Identifier" || object.type === "SafeMemberExpression") {
          left = object;
        } else {
          var ref = path.scope.generateDeclaredUidIdentifier("ref");
          node.object = ref;
          left = t.assignmentExpression("=", ref, object);
        }

        var nullCheck = t.binaryExpression("==", left, t.nullLiteral());
        node.type = "MemberExpression";
        path.replaceWith(node);

        // Gather trailing subscripts/calls, which are parent nodes:
        // eg; in `o?.x.y()`, group trailing `.x.y()` into the ternary
        var tail = path;
        while (tail.parentPath) {
          var parent = tail.parentPath;
          var hasChainedParent = parent.isMemberExpression() || parent.isCallExpression() && parent.get("callee") === tail || parent.node.type === "TildeCallExpression" && parent.get("left") === tail;

          if (hasChainedParent) {
            tail = tail.parentPath;
          } else {
            break;
          }
        }

        var ternary = t.conditionalExpression(nullCheck, t.nullLiteral(), tail.node);
        tail.replaceWith(ternary);
      },

      AwaitExpression(path) {
        if (path.get("argument").isArrayExpression() || path.node.argument.type === "ArrayComprehension") {
          var promiseDotAllCall = t.callExpression(t.memberExpression(t.identifier("Promise"), t.identifier("all")), [path.node.argument]);
          path.get("argument").replaceWith(promiseDotAllCall);
        }
      },

      VariableDeclaration(path) {
        // Error on auto-const when shadowing variable
        if (path.node.kind === "const") {
          if (path.node.extra && path.node.extra.implicit === true) {
            var id = (0, _variables.getShadowingIdentifiers)(path)[0];
            if (id) {
              throw path.buildCodeFrameError(`\`${id.name}\` is shadowed from a higher scope. ` + `If you want to reassign the variable, use \`now ${id.name} = ...\`. ` + "If you want to declare a new shadowed \`const\` variable, " + `you must use \`const ${id.name} = ...\` explicitly.`);
            }
          }
        }
      },

      // collect functions to be imported for the stdlib
      ReferencedIdentifier(path) {
        if (stdlib === false) return;

        var node = path.node,
            scope = path.scope;

        if (stdlib[node.name] && !scope.hasBinding(node.name)) {
          collectStdlibImport(stdlib, imports, node.name);
        }
      },

      ExistentialExpression(path) {
        (0, _existential.transformExistentialExpression)(path);
      },

      MatchExpression(path) {
        (0, _match.transformMatchExpression)(path);
      },

      MatchStatement(path) {
        (0, _match.transformMatchStatement)(path);
      }

    });

    insertStdlibImports(path, imports, useRequire);
  }

  return {
    manipulateOptions(opts, parserOpts, file) {
      if (!shouldParseAsLightScript(file)) return;

      opts.parserOpts = opts.parserOpts || {};
      opts.parserOpts.parser = _babylonLightscript.parse;
      parserOpts.plugins.unshift("lightscript");
      // TODO: allow configuration options to disable these, as they slow down parsing
      parserOpts.plugins.push("jsx", "flow");
    },

    visitor: {
      Program
    }

  };
};

var _babylonLightscript = require("@oigroup/babylon-lightscript");

var _stdlib = require("./stdlib");

var _babelInstance = require("@oigroup/lightscript-ast-transforms/lib/babelInstance");

var _is = require("@oigroup/lightscript-ast-transforms/lib/is");

var _is2 = _interopRequireDefault(_is);

var _functions = require("@oigroup/lightscript-ast-transforms/lib/functions");

var _blocks = require("@oigroup/lightscript-ast-transforms/lib/blocks");

var _variables = require("@oigroup/lightscript-ast-transforms/lib/variables");

var _classes = require("@oigroup/lightscript-ast-transforms/lib/classes");

var _match = require("@oigroup/lightscript-ast-transforms/lib/match");

var _safeCall = require("@oigroup/lightscript-ast-transforms/lib/safeCall");

var _comprehension = require("@oigroup/lightscript-ast-transforms/lib/comprehension");

var _existential = require("@oigroup/lightscript-ast-transforms/lib/existential");

var _inlinedOperators = require("@oigroup/lightscript-ast-transforms/lib/inlinedOperators");

var _for = require("@oigroup/lightscript-ast-transforms/lib/for");

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _toConsumableArray(arr) { if (Array.isArray(arr)) { for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) { arr2[i] = arr[i]; } return arr2; } else { return Array.from(arr); } }

module.exports = exports["default"];